@@ -10,6 +10,7 @@
            , ScopedTypeVariables
            , InstanceSigs
            , Rank2Types
+           , CPP
  #-}
 -- | This is an internal module, anything exported by this module
 -- may change without a major version bump.  Please use only
@@ -17,17 +18,14 @@
 module Database.Esqueleto.Internal.Sql
   ( -- * The pretty face
     SqlQuery
-  , SqlExpr
+  , SqlExpr(..)
   , SqlEntity
   , select
   , selectSource
-  , selectDistinct
-  , selectDistinctSource
   , delete
   , deleteCount
   , update
   , updateCount
-  , insertSelectDistinct
   , insertSelect
   , insertSelectCount
     -- * The guts
@@ -35,24 +33,33 @@ module Database.Esqueleto.Internal.Sql
   , unsafeSqlBinOp
   , unsafeSqlBinOpComposite
   , unsafeSqlValue
+  , unsafeSqlCastAs
   , unsafeSqlFunction
   , unsafeSqlExtractSubField
   , UnsafeSqlFunctionArgument
+  , OrderByClause
   , rawSelectSource
   , runSource
   , rawEsqueleto
   , toRawSql
   , Mode(..)
+  , NeedParens(..)
   , IdentState
   , initialIdentState
   , IdentInfo
   , SqlSelect(..)
   , veryUnsafeCoerceSqlExprValue
   , veryUnsafeCoerceSqlExprValueList
+  -- * Helper functions
+  , makeOrderByNoNewline
+  , uncommas'
+  , parens
+  , toArgList
+  , builderToText
   ) where
 
 import Control.Arrow ((***), first)
-import Control.Exception (throw, throwIO)
+import Control.Exception (Exception, throw, throwIO)
 import Control.Monad (ap, MonadPlus(..), void)
 import Control.Monad.IO.Class (MonadIO(..))
 import Control.Monad.Trans.Class (lift)
@@ -60,7 +67,10 @@ import Control.Monad.Trans.Resource (MonadResource, release)
 import Data.Acquire (with, allocateAcquire, Acquire)
 import Data.Int (Int64)
 import Data.List (intersperse)
-import Data.Monoid (Last(..), (<>))
+#if __GLASGOW_HASKELL__ < 804
+import Data.Semigroup
+#endif
+import qualified Data.Monoid as Monoid
 import Data.Proxy (Proxy(..))
 import Database.Esqueleto.Internal.PersistentImport
 import Database.Persist.Sql.Util (entityColumnNames, entityColumnCount, parseEntityValues, isIdField, hasCompositeKey)
@@ -76,6 +86,47 @@ import qualified Data.Text.Lazy.Builder as TLB
 
 import Database.Esqueleto.Internal.Language
 
+-- | Exception data type for @esqueleto@ internal errors
+data EsqueletoError =
+    CompositeKeyErr CompositeKeyError
+  | UnexpectedCaseErr UnexpectedCaseError
+  | SqlBinOpCompositeErr SqlBinOpCompositeError
+  deriving (Show)
+
+instance Exception EsqueletoError
+
+data CompositeKeyError =
+    NotError
+  | ToInsertionError
+  | CombineInsertionError
+  | FoldHelpError
+  | SqlCaseError
+  | SqlBinOpError
+  | SqlCastAsError
+  | MakeOnClauseError
+  | MakeExcError
+  | MakeSetError
+  | MakeWhereError
+  | MakeHavingError
+  deriving (Show)
+
+data UnexpectedCaseError =
+    EmptySqlExprValueList
+  | MakeFromError
+  | UnsupportedSqlInsertIntoType
+  | InsertionFinalError
+  | NewIdentForError
+  | UnsafeSqlCaseError
+  deriving (Show)
+
+data SqlBinOpCompositeError =
+    MismatchingLengthsError
+  | NullPlaceholdersError
+  | DeconstructionError
+  deriving (Show)
+
+
+
 -- | SQL backend for @esqueleto@ using 'SqlPersistT'.
 newtype SqlQuery a =
   Q { unQ :: W.WriterT SideData (S.State IdentState) a }
@@ -112,11 +163,13 @@ data SideData = SideData { sdDistinctClause :: !DistinctClause
                          , sdLockingClause  :: !LockingClause
                          }
 
-instance Monoid SideData where
-  mempty = SideData mempty mempty mempty mempty mempty mempty mempty mempty mempty
-  SideData d f s w g h o l k `mappend` SideData d' f' s' w' g' h' o' l' k' =
+instance Semigroup SideData where
+  SideData d f s w g h o l k <> SideData d' f' s' w' g' h' o' l' k' =
     SideData (d <> d') (f <> f') (s <> s') (w <> w') (g <> g') (h <> h') (o <> o') (l <> l') (k <> k')
 
+instance Monoid SideData where
+  mempty = SideData mempty mempty mempty mempty mempty mempty mempty mempty mempty
+  mappend = (<>)
 
 -- | The @DISTINCT@ "clause".
 data DistinctClause =
@@ -124,13 +177,15 @@ data DistinctClause =
   | DistinctStandard                -- ^ Only @DISTINCT@, SQL standard.
   | DistinctOn [SqlExpr DistinctOn] -- ^ @DISTINCT ON@, PostgreSQL extension.
 
+instance Semigroup DistinctClause where
+  DistinctOn a     <> DistinctOn b = DistinctOn (a <> b)
+  DistinctOn a     <> _            = DistinctOn a
+  DistinctStandard <> _            = DistinctStandard
+  DistinctAll      <> b            = b
+
 instance Monoid DistinctClause where
   mempty = DistinctAll
-  DistinctOn a     `mappend` DistinctOn b = DistinctOn (a <> b)
-  DistinctOn a     `mappend` _            = DistinctOn a
-  DistinctStandard `mappend` _            = DistinctStandard
-  DistinctAll      `mappend` b            = b
-
+  mappend = (<>)
 
 -- | A part of a @FROM@ clause.
 data FromClause =
@@ -177,19 +232,24 @@ collectOnClauses = go []
 data WhereClause = Where (SqlExpr (Value Bool))
                  | NoWhere
 
+instance Semigroup WhereClause where
+  NoWhere  <> w        = w
+  w        <> NoWhere  = w
+  Where e1 <> Where e2 = Where (e1 &&. e2)
+
 instance Monoid WhereClause where
   mempty = NoWhere
-  NoWhere  `mappend` w        = w
-  w        `mappend` NoWhere  = w
-  Where e1 `mappend` Where e2 = Where (e1 &&. e2)
-
+  mappend = (<>)
 
 -- | A @GROUP BY@ clause.
 newtype GroupByClause = GroupBy [SomeValue SqlExpr]
 
+instance Semigroup GroupByClause where
+  GroupBy fs <> GroupBy fs' = GroupBy (fs <> fs')
+
 instance Monoid GroupByClause where
   mempty = GroupBy []
-  GroupBy fs `mappend` GroupBy fs' = GroupBy (fs <> fs')
+  mappend = (<>)
 
 -- | A @HAVING@ cause.
 type HavingClause = WhereClause
@@ -201,17 +261,19 @@ type OrderByClause = SqlExpr OrderBy
 -- | A @LIMIT@ clause.
 data LimitClause = Limit (Maybe Int64) (Maybe Int64)
 
-instance Monoid LimitClause where
-  mempty = Limit mzero mzero
-  Limit l1 o1 `mappend` Limit l2 o2 =
+instance Semigroup LimitClause where
+  Limit l1 o1 <> Limit l2 o2 =
     Limit (l2 `mplus` l1) (o2 `mplus` o1)
     -- More than one 'limit' or 'offset' is issued, we want to
     -- keep the latest one.  That's why we use mplus with
     -- "reversed" arguments.
 
+instance Monoid LimitClause where
+  mempty = Limit mzero mzero
+  mappend = (<>)
 
 -- | A locking clause.
-type LockingClause = Last LockingKind
+type LockingClause = Monoid.Last LockingKind
 
 
 ----------------------------------------------------------------------
@@ -238,7 +300,7 @@ newIdentFor = Q . lift . try . unDBName
       s <- S.get
       let go (t:ts) | t `HS.member` inUse s = go ts
                     | otherwise             = use t
-          go [] = error "Esqueleto/Sql/newIdentFor: never here"
+          go [] = throw (UnexpectedCaseErr NewIdentForError)
       go (possibilities orig)
 
     possibilities t = t : map addNum [2..]
@@ -260,7 +322,6 @@ useIdent :: IdentInfo -> Ident -> TLB.Builder
 useIdent info (I ident) = fromDBName info $ DBName ident
 
 
-----------------------------------------------------------------------
 
 
 -- | An expression on the SQL backend.
@@ -395,7 +456,7 @@ instance Esqueleto SqlQuery SqlExpr SqlBackend where
 
   having expr = Q $ W.tell mempty { sdHavingClause = Where expr }
 
-  locking kind = Q $ W.tell mempty { sdLockingClause = Last (Just kind) }
+  locking kind = Q $ W.tell mempty { sdLockingClause = Monoid.Last (Just kind) }
 
   orderBy exprs = Q $ W.tell mempty { sdOrderByClause = exprs }
   asc  = EOrderBy ASC
@@ -416,9 +477,10 @@ instance Esqueleto SqlQuery SqlExpr SqlBackend where
     where
       toDistinctOn :: SqlExpr OrderBy -> SqlExpr DistinctOn
       toDistinctOn (EOrderBy _ f) = EDistinctOn f
+      toDistinctOn EOrderRandom =
+        error "We can't select distinct by a random order!"
 
   sub_select         = sub SELECT
-  sub_selectDistinct = sub_select . distinct
 
   (^.) :: forall val typ. (PersistEntity val, PersistField typ)
        => SqlExpr (Entity val) -> EntityField val typ -> SqlExpr (Value typ)
@@ -431,6 +493,14 @@ instance Esqueleto SqlQuery SqlExpr SqlBackend where
       ed          = entityDef $ getEntityVal (Proxy :: Proxy (SqlExpr (Entity val)))
       Just pdef   = entityPrimary ed
 
+  withNonNull :: PersistField typ
+              => SqlExpr (Value (Maybe typ))
+              -> (SqlExpr (Value typ) -> SqlQuery a)
+              -> SqlQuery a
+  withNonNull field f = do
+    where_ $ not_ $ isNothing field
+    f $ veryUnsafeCoerceSqlExprValue field
+
   EMaybe r ?. field = just (r ^. field)
 
   val v = ERaw Never $ const ("?", [toPersistValue v])
@@ -448,7 +518,7 @@ instance Esqueleto SqlQuery SqlExpr SqlBackend where
 
   not_ (ERaw p f) = ERaw Never $ \info -> let (b, vals) = f info
                                           in ("NOT " <> parensM p b, vals)
-  not_ (ECompositeKey _) = unexpectedCompositeKeyError "not_"
+  not_ (ECompositeKey _) = throw (CompositeKeyErr NotError)
 
   (==.) = unsafeSqlBinOpComposite " = " " AND "
   (!=.) = unsafeSqlBinOpComposite " != " " OR "
@@ -488,7 +558,6 @@ instance Esqueleto SqlQuery SqlExpr SqlBackend where
   castString = veryUnsafeCoerceSqlExprValue
 
   subList_select         = EList . sub_select
-  subList_selectDistinct = subList_select . distinct
 
   valList []   = EEmptyList
   valList vals = EList $ ERaw Parens $ const ( uncommas ("?" <$ vals)
@@ -514,13 +583,13 @@ instance Esqueleto SqlQuery SqlExpr SqlBackend where
   field /=. expr = setAux field (\ent -> ent ^. field /. expr)
 
   (<#) _ (ERaw _ f)        = EInsert Proxy f
-  (<#) _ (ECompositeKey _) = unexpectedCompositeKeyError "(<#)"
+  (<#) _ (ECompositeKey _) = throw (CompositeKeyErr ToInsertionError)
 
   (EInsert _ f) <&> (ERaw _ g) = EInsert Proxy $ \x ->
     let (fb, fv) = f x
         (gb, gv) = g x
     in (fb <> ", " <> gb, fv ++ gv)
-  (EInsert _ _) <&> (ECompositeKey _) = unexpectedCompositeKeyError "(<&>)"
+  (EInsert _ _) <&> (ECompositeKey _) = throw (CompositeKeyErr CombineInsertionError)
 
   case_ = unsafeSqlCase
   toBaseId = veryUnsafeCoerceSqlExprValue
@@ -578,7 +647,7 @@ unsafeSqlCase when (ERaw p1 f1) = ERaw Never buildCase
         in ( "CASE" <> b2 <> " ELSE " <> parensM p1 b1 <> " END", vals2 <> vals1)
 
     mapWhen :: [(SqlExpr (Value Bool), SqlExpr (Value a))] -> IdentInfo -> (TLB.Builder, [PersistValue])
-    mapWhen []    _    = error "unsafeSqlCase: empty when list."
+    mapWhen []    _    = throw (UnexpectedCaseErr UnsafeSqlCaseError)
     mapWhen when' info = foldl (foldHelp info) (mempty, mempty) when'
 
     foldHelp :: IdentInfo -> (TLB.Builder, [PersistValue]) -> (SqlExpr (Value Bool), SqlExpr (Value a)) -> (TLB.Builder, [PersistValue])
@@ -586,8 +655,8 @@ unsafeSqlCase when (ERaw p1 f1) = ERaw Never buildCase
         let (b1, vals1) = f1' info
             (b2, vals2) = f2 info
         in ( b0 <> " WHEN " <> parensM p1' b1 <> " THEN " <> parensM p2 b2, vals0 <> vals1 <> vals2 )
-    foldHelp _ _ _ = unexpectedCompositeKeyError "unsafeSqlCase/foldHelp"
-unsafeSqlCase _ (ECompositeKey _) = unexpectedCompositeKeyError "unsafeSqlCase"
+    foldHelp _ _ _ = throw (CompositeKeyErr FoldHelpError)
+unsafeSqlCase _ (ECompositeKey _) = throw (CompositeKeyErr SqlCaseError)
 
 
 -- | (Internal) Create a custom binary operator.  You /should/
@@ -609,7 +678,7 @@ unsafeSqlBinOp op (ERaw p1 f1) (ERaw p2 f2) = ERaw Parens f
                  (b2, vals2) = f2 info
              in ( parensM p1 b1 <> op <> parensM p2 b2
                 , vals1 <> vals2 )
-unsafeSqlBinOp _ _ _ = unexpectedCompositeKeyError "unsafeSqlBinOp"
+unsafeSqlBinOp _ _ _ = throw (CompositeKeyErr SqlBinOpError)
 {-# INLINE unsafeSqlBinOp #-}
 
 
@@ -647,20 +716,17 @@ unsafeSqlBinOpComposite op sep a b = ERaw Parens $ compose (listify a) (listify
     deconstruct :: (TLB.Builder, [PersistValue]) -> ([TLB.Builder], [PersistValue])
     deconstruct ("?", [PersistList vals]) = (replicate (length vals) "?", vals)
     deconstruct (b', []) = (TLB.fromLazyText <$> TL.splitOn "," (TLB.toLazyText b'), [])
-    deconstruct x = err $ "cannot deconstruct " ++ show x ++ "."
+    deconstruct _ = throw (SqlBinOpCompositeErr DeconstructionError)
 
     compose f1 f2 info
-      | not (null v1 || null v2) = err' "one side needs to have null placeholders"
-      | length b1 /= length b2   = err' "mismatching lengths"
+      | not (null v1 || null v2) = throw (SqlBinOpCompositeErr NullPlaceholdersError)
+      | length b1 /= length b2   = throw (SqlBinOpCompositeErr MismatchingLengthsError)
       | otherwise                = (bc, vc)
       where
         (b1, v1) = f1 info
         (b2, v2) = f2 info
         bc = intersperseB sep [x <> op <> y | (x, y) <- zip b1 b2]
         vc = v1 <> v2
-        err' = err . (++ (", " ++ show ((b1, v1), (b2, v2))))
-
-    err = error . ("unsafeSqlBinOpComposite: " ++)
 
 -- | (Internal) A raw SQL value.  The same warning from
 -- 'unsafeSqlBinOp' applies to this function as well.
@@ -701,6 +767,14 @@ unsafeSqlFunctionParens name arg =
           uncommas' $ map (\(ERaw p f) -> first (parensM p) (f info)) $ toArgList arg
     in (name <> parens argsTLB, argsVals)
 
+-- | (Internal) An explicit SQL type cast using CAST(value as type).
+-- See 'unsafeSqlBinOp' for warnings.
+unsafeSqlCastAs :: T.Text -> SqlExpr (Value a) -> SqlExpr (Value b)
+unsafeSqlCastAs t (ERaw p f) =
+  ERaw Never $ \info ->
+    let (b, v) = f info
+    in ("CAST" <> parens ( parensM p b <> " AS " <> TLB.fromText t), v )
+unsafeSqlCastAs _ (ECompositeKey _) = throw (CompositeKeyErr SqlCastAsError)
 
 class UnsafeSqlFunctionArgument a where
   toArgList :: a -> [SqlExpr (Value ())]
@@ -739,26 +813,26 @@ veryUnsafeCoerceSqlExprValue (ECompositeKey f) = ECompositeKey f
 -- a)' to 'SqlExpr (Value a)'.  Does not work with empty lists.
 veryUnsafeCoerceSqlExprValueList :: SqlExpr (ValueList a) -> SqlExpr (Value a)
 veryUnsafeCoerceSqlExprValueList (EList v)  = v
-veryUnsafeCoerceSqlExprValueList EEmptyList =
-  error "veryUnsafeCoerceSqlExprValueList: empty list."
+veryUnsafeCoerceSqlExprValueList EEmptyList = throw (UnexpectedCaseErr EmptySqlExprValueList)
 
 
 ----------------------------------------------------------------------
 
-
 -- | (Internal) Execute an @esqueleto@ @SELECT@ 'SqlQuery' inside
 -- @persistent@'s 'SqlPersistT' monad.
 rawSelectSource :: ( SqlSelect a r
                    , MonadIO m1
-                   , MonadIO m2 )
+                   , MonadIO m2
+                   )
                  => Mode
                  -> SqlQuery a
-                 -> SqlReadT m1 (Acquire (C.Source m2 r))
+                 -> SqlReadT m1 (Acquire (C.ConduitT () r m2 ()))
 rawSelectSource mode query =
       do
-        conn <- persistBackend <$> R.ask
-        res <- run conn
-        return $ (C.$= massage) `fmap` res
+        conn <- projectBackend <$> R.ask
+        let _ = conn :: SqlBackend
+        res <- R.withReaderT (const conn) (run conn)
+        return $ (C..| massage) `fmap` res
     where
 
       run conn =
@@ -779,9 +853,13 @@ rawSelectSource mode query =
 -- | Execute an @esqueleto@ @SELECT@ query inside @persistent@'s
 -- 'SqlPersistT' monad and return a 'C.Source' of rows.
 selectSource :: ( SqlSelect a r
-                , MonadResource m )
+               , BackendCompatible SqlBackend backend
+               , IsPersistBackend backend
+               , PersistQueryRead backend
+               , PersistStoreRead backend, PersistUniqueRead backend
+               , MonadResource m )
              => SqlQuery a
-             -> C.Source (SqlPersistT m) r
+             -> C.ConduitT () r (R.ReaderT backend m) ()
 selectSource query = do
   res <- lift $ rawSelectSource SELECT query
   (key, src) <- lift $ allocateAcquire res
@@ -830,40 +908,19 @@ selectSource query = do
 -- function composition that the @p@ inside the query is of type
 -- @SqlExpr (Entity Person)@.
 select :: ( SqlSelect a r
-          , MonadIO m )
+          , MonadIO m
+          )
        => SqlQuery a -> SqlReadT m [r]
 select query = do
     res <- rawSelectSource SELECT query
     conn <- R.ask
     liftIO $ with res $ flip R.runReaderT conn . runSource
 
-
--- | Execute an @esqueleto@ @SELECT DISTINCT@ query inside
--- @persistent@'s 'SqlPersistT' monad and return a 'C.Source' of
--- rows.
-selectDistinctSource
-  :: ( SqlSelect a r
-     , MonadResource m )
-  => SqlQuery a
-  -> C.Source (SqlPersistT m) r
-selectDistinctSource = selectSource . distinct
-{-# DEPRECATED selectDistinctSource "Since 2.2.4: use 'selectSource' and 'distinct'." #-}
-
-
--- | Execute an @esqueleto@ @SELECT DISTINCT@ query inside
--- @persistent@'s 'SqlPersistT' monad and return a list of rows.
-selectDistinct :: ( SqlSelect a r
-                  , MonadIO m )
-               => SqlQuery a -> SqlPersistT m [r]
-selectDistinct = select . distinct
-{-# DEPRECATED selectDistinct "Since 2.2.4: use 'select' and 'distinct'." #-}
-
-
 -- | (Internal) Run a 'C.Source' of rows.
 runSource :: Monad m =>
-             C.Source (R.ReaderT backend m) r
+             C.ConduitT () r (R.ReaderT backend m) ()
           -> R.ReaderT backend m [r]
-runSource src = src C.$$ CL.consume
+runSource src = C.runConduit $ src C..| CL.consume
 
 
 ----------------------------------------------------------------------
@@ -871,12 +928,12 @@ runSource src = src C.$$ CL.consume
 
 -- | (Internal) Execute an @esqueleto@ statement inside
 -- @persistent@'s 'SqlPersistT' monad.
-rawEsqueleto :: ( MonadIO m, SqlSelect a r, IsSqlBackend backend)
+rawEsqueleto :: ( MonadIO m, SqlSelect a r, BackendCompatible SqlBackend backend)
            => Mode
            -> SqlQuery a
            -> R.ReaderT backend m Int64
 rawEsqueleto mode query = do
-  conn <- persistBackend <$> R.ask
+  conn <- R.ask
   uncurry rawExecuteCount $
     first builderToText $
     toRawSql mode (conn, initialIdentState) query
@@ -928,17 +985,23 @@ deleteCount = rawEsqueleto DELETE
 -- 'set' p [ PersonAge '=.' 'just' ('val' thisYear) -. p '^.' PersonBorn ]
 -- 'where_' $ isNothing (p '^.' PersonAge)
 -- @
-update :: ( MonadIO m
-          , SqlEntity val )
-       => (SqlExpr (Entity val) -> SqlQuery ())
-       -> SqlWriteT m ()
+update
+  ::
+  ( MonadIO m, PersistEntity val
+  , BackendCompatible SqlBackend (PersistEntityBackend val)
+  )
+  => (SqlExpr (Entity val) -> SqlQuery ())
+  -> SqlWriteT m ()
 update = void . updateCount
 
 -- | Same as 'update', but returns the number of rows affected.
-updateCount :: ( MonadIO m
-               , SqlEntity val )
-            => (SqlExpr (Entity val) -> SqlQuery ())
-            -> SqlWriteT m Int64
+updateCount
+  ::
+  ( MonadIO m, PersistEntity val
+  , BackendCompatible SqlBackend (PersistEntityBackend val)
+  )
+  => (SqlExpr (Entity val) -> SqlQuery ())
+  -> SqlWriteT m Int64
 updateCount = rawEsqueleto UPDATE . from
 
 
@@ -958,7 +1021,7 @@ builderToText = TL.toStrict . TLB.toLazyTextWith defaultChunkSize
 -- possible but tedious), you may just turn on query logging of
 -- @persistent@.
 toRawSql
-  :: (IsSqlBackend backend, SqlSelect a r)
+  :: (SqlSelect a r, BackendCompatible SqlBackend backend)
   => Mode -> (backend, IdentState) -> SqlQuery a -> (TLB.Builder, [PersistValue])
 toRawSql mode (conn, firstIdentState) query =
   let ((ret, sd), finalIdentState) =
@@ -978,7 +1041,7 @@ toRawSql mode (conn, firstIdentState) query =
       -- that were used) to the subsequent calls.  This ensures
       -- that no name clashes will occur on subqueries that may
       -- appear on the expressions below.
-      info = (persistBackend conn, finalIdentState)
+      info = (projectBackend conn, finalIdentState)
   in mconcat
       [ makeInsertInto info mode ret
       , makeSelect     info mode distinctClause ret
@@ -1055,7 +1118,7 @@ makeFrom info mode fs = ret
               , mk Parens rhs
               , maybe mempty makeOnClause monClause
               ]
-    mk _ (OnClause _) = error "Esqueleto/Sql/makeFrom: never here (is collectOnClauses working?)"
+    mk _ (OnClause _) = throw (UnexpectedCaseErr MakeFromError)
 
     base ident@(I identText) def =
       let db@(DBName dbText) = entityDB def
@@ -1071,29 +1134,25 @@ makeFrom info mode fs = ret
     fromKind FullOuterJoinKind  = " FULL OUTER JOIN "
 
     makeOnClause (ERaw _ f)        = first (" ON " <>) (f info)
-    makeOnClause (ECompositeKey _) = unexpectedCompositeKeyError "makeFrom/makeOnClause"
+    makeOnClause (ECompositeKey _) = throw (CompositeKeyErr MakeOnClauseError)
 
     mkExc :: SqlExpr (Value Bool) -> OnClauseWithoutMatchingJoinException
     mkExc (ERaw _ f) =
       OnClauseWithoutMatchingJoinException $
       TL.unpack $ TLB.toLazyText $ fst (f info)
-    mkExc (ECompositeKey _) = unexpectedCompositeKeyError "makeFrom/mkExc"
-
-unexpectedCompositeKeyError :: String -> a
-unexpectedCompositeKeyError w = error $ w ++ ": non-id/composite keys not expected here"
-
+    mkExc (ECompositeKey _) = throw (CompositeKeyErr MakeExcError)
 
 makeSet :: IdentInfo -> [SetClause] -> (TLB.Builder, [PersistValue])
 makeSet _    [] = mempty
 makeSet info os = first ("\nSET " <>) . uncommas' $ concatMap mk os
   where
     mk (SetClause (ERaw _ f))        = [f info]
-    mk (SetClause (ECompositeKey _)) = unexpectedCompositeKeyError "makeSet" -- FIXME
+    mk (SetClause (ECompositeKey _)) = throw (CompositeKeyErr MakeSetError) -- FIXME
 
 makeWhere :: IdentInfo -> WhereClause -> (TLB.Builder, [PersistValue])
 makeWhere _    NoWhere                   = mempty
 makeWhere info (Where (ERaw _ f))        = first ("\nWHERE " <>) (f info)
-makeWhere _    (Where (ECompositeKey _)) = unexpectedCompositeKeyError "makeWhere"
+makeWhere _    (Where (ECompositeKey _)) = throw (CompositeKeyErr MakeWhereError)
 
 
 makeGroupBy :: IdentInfo -> GroupByClause -> (TLB.Builder, [PersistValue])
@@ -1110,12 +1169,13 @@ makeGroupBy info (GroupBy fields) = first ("\nGROUP BY " <>) build
 makeHaving :: IdentInfo -> WhereClause -> (TLB.Builder, [PersistValue])
 makeHaving _    NoWhere                    = mempty
 makeHaving info (Where (ERaw _ f))         = first ("\nHAVING " <>) (f info)
-makeHaving _    (Where (ECompositeKey _)) = unexpectedCompositeKeyError "makeHaving"
+makeHaving _    (Where (ECompositeKey _)) = throw (CompositeKeyErr MakeHavingError)
 
 -- makeHaving, makeWhere and makeOrderBy
-makeOrderBy :: IdentInfo -> [OrderByClause] -> (TLB.Builder, [PersistValue])
-makeOrderBy _    [] = mempty
-makeOrderBy info os = first ("\nORDER BY " <>) . uncommas' $ concatMap mk os
+makeOrderByNoNewline ::
+     IdentInfo -> [OrderByClause] -> (TLB.Builder, [PersistValue])
+makeOrderByNoNewline _    [] = mempty
+makeOrderByNoNewline info os = first ("ORDER BY " <>) . uncommas' $ concatMap mk os
   where
     mk :: OrderByClause -> [(TLB.Builder, [PersistValue])]
     mk (EOrderBy t (ERaw p f)) = [first ((<> orderByType t) . parensM p) (f info)]
@@ -1127,6 +1187,13 @@ makeOrderBy info os = first ("\nORDER BY " <>) . uncommas' $ concatMap mk os
     orderByType ASC  = " ASC"
     orderByType DESC = " DESC"
 
+makeOrderBy :: IdentInfo -> [OrderByClause] -> (TLB.Builder, [PersistValue])
+makeOrderBy _ [] = mempty
+makeOrderBy info is =
+  let (tlb, vals) = makeOrderByNoNewline info is
+  in ("\n" <> tlb, vals)
+
+{-# DEPRECATED EOrderRandom "Since 2.6.0: `rand` ordering function is not uniform across all databases! To avoid accidental partiality it will be removed in the next major version." #-}
 
 makeLimit :: IdentInfo -> LimitClause -> [OrderByClause] -> (TLB.Builder, [PersistValue])
 makeLimit (conn, _) (Limit ml mo) orderByClauses =
@@ -1137,11 +1204,12 @@ makeLimit (conn, _) (Limit ml mo) orderByClauses =
 
 
 makeLocking :: LockingClause -> (TLB.Builder, [PersistValue])
-makeLocking = flip (,) [] . maybe mempty toTLB . getLast
+makeLocking = flip (,) [] . maybe mempty toTLB . Monoid.getLast
   where
-    toTLB ForUpdate       = "\nFOR UPDATE"
-    toTLB ForShare        = "\nFOR SHARE"
-    toTLB LockInShareMode = "\nLOCK IN SHARE MODE"
+    toTLB ForUpdate           = "\nFOR UPDATE"
+    toTLB ForUpdateSkipLocked = "\nFOR UPDATE SKIP LOCKED"
+    toTLB ForShare            = "\nFOR SHARE"
+    toTLB LockInShareMode     = "\nLOCK IN SHARE MODE"
 
 
 
@@ -1172,7 +1240,7 @@ class SqlSelect a r | a -> r, r -> a where
 
   -- | Create @INSERT INTO@ clause instead.
   sqlInsertInto :: IdentInfo -> a -> (TLB.Builder, [PersistValue])
-  sqlInsertInto = error "Type does not support sqlInsertInto."
+  sqlInsertInto = throw (UnexpectedCaseErr UnsupportedSqlInsertIntoType)
 
 
 -- | @INSERT INTO@ hack.
@@ -1186,9 +1254,8 @@ instance SqlSelect (SqlExpr InsertFinal) InsertFinal where
     in ("INSERT INTO " <> table <> parens fields <> "\n", [])
   sqlSelectCols info (EInsertFinal (EInsert _ f)) = f info
   sqlSelectColCount   = const 0
-  sqlSelectProcessRow = const (Right (error msg))
-    where
-      msg = "sqlSelectProcessRow/SqlSelect/InsertionFinal: never here"
+  sqlSelectProcessRow =
+    const (Right (throw (UnexpectedCaseErr InsertionFinalError)))
 
 
 -- | Not useful for 'select', but used for 'update' and 'delete'.
@@ -1755,10 +1822,3 @@ insertSelect = void . insertSelectCount
 insertSelectCount :: (MonadIO m, PersistEntity a) =>
   SqlQuery (SqlExpr (Insertion a)) -> SqlWriteT m Int64
 insertSelectCount = rawEsqueleto INSERT_INTO . fmap EInsertFinal
-
-
--- | Insert a 'PersistField' for every unique selected value.
-insertSelectDistinct :: (MonadIO m, PersistEntity a) =>
-  SqlQuery (SqlExpr (Insertion a)) -> SqlWriteT m ()
-insertSelectDistinct = insertSelect . distinct
-{-# DEPRECATED insertSelectDistinct "Since 2.2.4: use 'insertSelect' and 'distinct'." #-}
